// Copyright 2019 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include "iree/compiler/Dialect/HAL/Utils/TypeUtils.h"

#include "iree/compiler/Dialect/HAL/IR/HALOps.h"
#include "iree/compiler/Dialect/HAL/IR/HALTypes.h"
#include "iree/compiler/Dialect/IREE/IR/IREETypes.h"
#include "iree/compiler/Dialect/Shape/IR/Builders.h"
#include "iree/compiler/Dialect/Shape/IR/ShapeOps.h"
#include "llvm/ADT/Optional.h"
#include "llvm/Support/Debug.h"
#include "mlir/Dialect/StandardOps/IR/Ops.h"
#include "mlir/IR/Attributes.h"
#include "mlir/IR/Builders.h"
#include "mlir/IR/BuiltinTypes.h"
#include "mlir/Support/LogicalResult.h"
#include "mlir/Transforms/DialectConversion.h"

namespace mlir {
namespace iree_compiler {
namespace IREE {
namespace HAL {

Value align(Location loc, Value value, int64_t alignment, OpBuilder &builder) {
  // (value + (alignment - 1)) & ~(alignment - 1)
  return builder.createOrFold<AndOp>(
      loc,
      builder.createOrFold<AddIOp>(
          loc, value,
          builder.createOrFold<ConstantIndexOp>(loc, alignment - 1)),
      builder.createOrFold<ConstantIndexOp>(loc, ~(alignment - 1)));
}

int32_t getRoundedElementByteWidth(Type type) {
  return (type.getIntOrFloatBitWidth() + 8 - 1) / 8;
}

TensorType convertTensorTypeToABIType(TensorType sourceType) {
  assert(sourceType.hasRank() && "only ranked tensors are supported");
  Type sourceElementType = sourceType.getElementType();
  Type targetElementType = sourceElementType;
  if (auto sourceIntType = sourceElementType.dyn_cast<IntegerType>()) {
    int32_t targetByteWidth = getRoundedElementByteWidth(sourceElementType);
    targetElementType =
        IntegerType::get(sourceElementType.getContext(), targetByteWidth * 8);
  }
  return RankedTensorType::get(sourceType.getShape(), targetElementType);
}

SmallVector<Value, 4> getStaticShapeDims(Location loc, ShapedType shapedType,
                                         OpBuilder &builder) {
  SmallVector<Value, 4> shape;
  if (shapedType.getRank() >= 1) {
    for (auto dim : shapedType.getShape()) {
      shape.push_back(builder.createOrFold<mlir::ConstantIndexOp>(loc, dim));
    }
  }
  return shape;
}

llvm::Optional<SmallVector<Value, 4>> getShapeDims(Location loc,
                                                   Value shapedValue,
                                                   OpBuilder &builder) {
  ShapedType shapedType = shapedValue.getType().cast<ShapedType>();
  if (shapedType.hasStaticShape()) {
    return getStaticShapeDims(loc, shapedType, builder);
  } else {
    // Dynamic shape lookup.
    Value rsValue = Shape::buildOrFindRankedShapeForValue(
        loc, shapedValue, builder.getIndexType(), builder);
    if (!rsValue) {
      return llvm::None;
    }
    SmallVector<Value, 4> dims;
    // Note that in the following, we require that the dims resolve
    // to discrete SSA values, which in a stream, will be block args.
    if (failed(Shape::getRankedDimsFromRankedShape(
            loc, rsValue, /*createIntermediateOps=*/true, dims, builder))) {
      return llvm::None;
    }
    return dims;
  }
}

Value getValueSize(Location loc, Value value, OpBuilder &builder) {
  // Function arguments are special as we always have to query.
  auto definingOp = value.getDefiningOp();
  if (!definingOp) {
    return builder.createOrFold<IREE::HAL::BufferLengthOp>(
        loc, builder.getIndexType(), value);
  }

  if (auto awareOp = dyn_cast_or_null<SizeAwareOpInterface>(definingOp)) {
    return awareOp.getResultSizeFromValue(value);
  }

  auto type = value.getType();
  if (auto awareType = type.dyn_cast<SizeAwareTypeInterface>()) {
    auto sizeValue = awareType.getSize(value);
    if (sizeValue) return sizeValue;
  }
  if (auto inferType = type.dyn_cast<InferTypeSizeInterface>()) {
    return inferType.inferSizeFromValue(loc, value, builder);
  }

  auto elementType = IREE::HAL::getElementTypeValue(
      value.getType().cast<ShapedType>().getElementType());
  if (!elementType) return {};
  auto shape = IREE::HAL::getShapeDims(loc, value, builder);
  if (!shape) return {};
  auto deviceValue = builder.createOrFold<IREE::HAL::ExSharedDeviceOp>(loc);
  auto allocatorValue =
      builder.createOrFold<IREE::HAL::DeviceAllocatorOp>(loc, deviceValue);
  return builder.createOrFold<IREE::HAL::AllocatorComputeSizeOp>(
      loc, allocatorValue, *shape, elementType.getValue());
}

// static
bool TensorRewriteAdaptor::isValidNewType(Type newType) {
  return newType.isa<IREE::HAL::BufferType>() ||
         newType.isa<IREE::HAL::BufferViewType>();
}

// static
LogicalResult TensorRewriteAdaptor::verifyConstructionInvariants(
    Location loc, Value oldValue, Value newValue,
    ConversionPatternRewriter &rewriter) {
  if (!oldValue || !newValue) {
    return emitError(loc) << "TensorRewriteAdaptor values are null";
  }
  if (!oldValue.getType().isa<TensorType>()) {
    return emitError(loc) << "TensorRewriteAdaptor oldValue is not a Tensor";
  }
  if (!TensorRewriteAdaptor::isValidNewType(newValue.getType())) {
    return emitError(loc) << "TensorRewriteAdaptor newValue is invalid type "
                          << newValue.getType();
  }
  return success();
}

// static
TensorRewriteAdaptor TensorRewriteAdaptor::get(
    Location loc, Value oldValue, Value newValue,
    ConversionPatternRewriter &rewriter) {
  assert(succeeded(TensorRewriteAdaptor::verifyConstructionInvariants(
      loc, oldValue, newValue, rewriter)));
  return TensorRewriteAdaptor(loc, oldValue, newValue, rewriter);
}
// static
llvm::Optional<TensorRewriteAdaptor> TensorRewriteAdaptor::getChecked(
    Location loc, Value oldValue, Value newValue,
    ConversionPatternRewriter &rewriter) {
  if (failed(TensorRewriteAdaptor::verifyConstructionInvariants(
          loc, oldValue, newValue, rewriter))) {
    return llvm::None;
  }
  return TensorRewriteAdaptor(loc, oldValue, newValue, rewriter);
}

Value TensorRewriteAdaptor::getAllocator() {
  return rewriter_.createOrFold<IREE::HAL::BufferAllocatorOp>(
      loc_, AllocatorType::get(rewriter_.getContext()), getBuffer());
}

bool TensorRewriteAdaptor::isBufferView() {
  return newValue_.getType().isa<IREE::HAL::BufferViewType>();
}

Value TensorRewriteAdaptor::getBuffer() {
  if (isBufferView()) {
    return rewriter_.createOrFold<IREE::HAL::BufferViewBufferOp>(
        loc_, IREE::HAL::BufferType::get(rewriter_.getContext()), newValue_);
  } else {
    return newValue_;
  }
}

Value TensorRewriteAdaptor::getBufferView() {
  if (isBufferView()) {
    return newValue_;
  } else if (auto bufferViewBufferOp =
                 llvm::dyn_cast_or_null<IREE::HAL::BufferViewBufferOp>(
                     newValue_.getDefiningOp())) {
    return bufferViewBufferOp.buffer_view();
  } else {
    auto shapeDims = getShapeDims();
    if (!shapeDims) return {};
    return rewriter_.createOrFold<IREE::HAL::BufferViewCreateOp>(
        loc_, newValue_, getElementType(), *shapeDims);
  }
}

TensorType TensorRewriteAdaptor::getTensorType() {
  return oldValue_.getType().cast<TensorType>();
}

int32_t TensorRewriteAdaptor::getElementType() {
  return IREE::HAL::getElementTypeValue(getTensorType().getElementType())
      .getValueOr(0);
}

IntegerAttr TensorRewriteAdaptor::getElementTypeAttr() {
  return IREE::HAL::getElementTypeAttr(getTensorType().getElementType());
}

llvm::Optional<SmallVector<Value, 4>> TensorRewriteAdaptor::getShapeDims() {
  return IREE::HAL::getShapeDims(loc_, oldValue_, rewriter_);
}

llvm::Optional<SmallVector<Value, 4>> TensorRewriteAdaptor::getShapeDims(
    ConversionPatternRewriter &rewriter) {
  return IREE::HAL::getShapeDims(loc_, oldValue_, rewriter);
}

Value TensorRewriteAdaptor::getByteLength() {
  if (isBufferView()) {
    return rewriter_.createOrFold<IREE::HAL::BufferViewByteLengthOp>(
        loc_, getBufferView());
  } else {
    auto shapeDims = getShapeDims();
    if (!shapeDims) return {};
    return rewriter_.createOrFold<IREE::HAL::AllocatorComputeSizeOp>(
        loc_, getAllocator(), *shapeDims, getElementType());
  }
}

Value TensorRewriteAdaptor::computeOffset(ValueRange indices) {
  if (isBufferView()) {
    return rewriter_.createOrFold<IREE::HAL::BufferViewComputeOffsetOp>(
        loc_, getBufferView(), indices);
  } else {
    auto shapeDims = getShapeDims();
    if (!shapeDims) return {};
    return rewriter_.createOrFold<IREE::HAL::AllocatorComputeOffsetOp>(
        loc_, getAllocator(), *shapeDims, getElementType(), indices);
  }
}

llvm::Optional<TensorRewriteAdaptor::Range> TensorRewriteAdaptor::computeRange(
    ValueRange indices, ValueRange lengths) {
  if (isBufferView()) {
    auto range = rewriter_.create<IREE::HAL::BufferViewComputeRangeOp>(
        loc_, getBufferView(), indices, lengths);
    return Range{range.offset(), range.length()};
  } else {
    auto shapeDims = getShapeDims();
    if (!shapeDims) return llvm::None;
    auto range = rewriter_.create<IREE::HAL::AllocatorComputeRangeOp>(
        loc_, getAllocator(), *shapeDims, getElementType(), indices, lengths);
    return Range{range.offset(), range.length()};
  }
}

}  // namespace HAL
}  // namespace IREE
}  // namespace iree_compiler
}  // namespace mlir
